@@ -20,7 +20,8 @@ class D36pApiWrapperError(Exception):
 
 class BaseApiWrapper:
     def __init__(self, crmorg=None, token=None, tenant=None, client_id=None,
-                 client_secret=None, api_url='/api/data/v9.0/', extra_headers=None, page_size=100):
+                 client_secret=None, api_url='/api/data/v9.0/', extra_headers=None,
+                 page_size=100, impersonate=None):
         self.crmorg = crmorg or self.get_crmorg()
         self.tenant = tenant
         self.client_id = client_id
@@ -34,6 +35,7 @@ class BaseApiWrapper:
         self.current_page = None
         self.page_urls = {}
         self._api_url = api_url
+        self.impersonate = impersonate
         if extra_headers:
             self.__headers.update(extra_headers)
     entity_type = None
@@ -79,17 +81,14 @@ class BaseApiWrapper:
     def headers(self):
         headers = self.__headers
         headers.update({'Authorization': f'Bearer {self.token}'})
+        if self.impersonate:
+            headers.update({'CallerObjectId': self.impersonate})
         return headers
 
     @property
     def _base_api_url(self):
         return f"{self.crmorg}{self._api_url}"
 
-    @property
-    def has_next_page(self):
-        return self.current_page and self.page_urls.get(self.current_page + 1) \
-               and self.page_urls.get(self.current_page + 1) != '__none__'
-
     @property
     def api_url(self):
         if not self.entity_type:
@@ -113,18 +112,18 @@ class BaseApiWrapper:
         )
         return headers_dict
 
-    def get_page(self, page_url: str = None, select=None, request_filter=None, order_by=None, annotations=None, query_dict=None):
+    def get_page(self, page: int = 1, select=None, request_filter=None, order_by=None, annotations=None, query_dict=None):
         """
-        Used for getting a first page of list API data,
-        or a specific page if such page url has been cached.
+        Used for getting a particular page of list API data,
+        or a specific page. Will return empty "values" list if page is out of range
 
         :raises D365ApiError: if http request is not "Ok"
-        :raises D36pApiWrapperError: if requested page url is not cached or is out of range
-        :param page_url: autogenerated url for next page, returned with api response
+        :param page:
         :param select: string with comma separated attribute names of an entity to include, defaults to all
         :param request_filter: filter string
         :param order_by:
         :param annotations: if set to "*", will include all annotations for related objects
+        :param query_dict: arbitrary url params dictionary
         :return:
         """
         headers = self.headers.copy()
@@ -133,24 +132,36 @@ class BaseApiWrapper:
         )
         if annotations:
             headers = self._update_prefer_header(headers, 'odata.include-annotations', f'"{annotations}"')
-        params = {}
-        if not page_url:
-            page_url = self.api_url
-            if select:
-                params['$select'] = select
-            if request_filter:
-                params['$filter'] = request_filter
-            if order_by:
-                params['$orderby'] = order_by
-            if query_dict:
-                params.update(query_dict)
+        params = {
+            '$count': 'true'
+        }
+        if self._api_url.startswith('/api/data/v'):
+            params['$skiptoken'] = f'<cookie pagenumber="{page}" istracking="False" />'
+        elif page > 1:
+            params['$top'] = self.page_size
+            params['$skip'] = self.page_size * (page - 1)
+        if select:
+            params['$select'] = select
+        if request_filter:
+            params['$filter'] = request_filter
+        if order_by:
+            params['$orderby'] = order_by
+        if query_dict:
+            params.update(query_dict)
 
-        response = requests.get(page_url, params=params, headers=headers)
+        response = requests.get(self.api_url, params=params, headers=headers)
         if response.ok:
+            self.current_page = page
             return response.json()
         else:
             raise D365ApiError({'response_status_code': response.status_code, 'response_data': response.content})
 
+    def get_next_page(self):
+        if self.current_page:
+            return self.get_page(page=self.current_page + 1)
+        raise D36pApiWrapperError('To call "get_next_page()" "current_page" attribute can not be None. '
+                                  'Please call "get_page()" method first.')
+
     def get_previous_page(self):
         if not self.current_page or self.current_page == 1:
             raise D36pApiWrapperError('To call "get_next_page()" "current_page" attribute can not be None or 1. Please '
